from pathlib import Path
from textwrap import dedent

import pytest
from pytest import MonkeyPatch

from anaconda_auth.config import AnacondaAuthConfig

conda = pytest.importorskip("conda")


@pytest.mark.usefixtures("disable_dot_env")
def test_conda_context_apply_to_default_site(
    condarc_path: Path,
    config_toml: Path,
) -> None:
    condarc_path.write_text(
        dedent("""\
        ssl_verify: false
        proxy_servers:
            http: 127.0.0.1:80
            https: 127.0.0.1:80
    """)
    )
    conda.base.context.reset_context()

    config = AnacondaAuthConfig()
    assert config.proxy_servers == {"http": "127.0.0.1:80", "https": "127.0.0.1:80"}
    assert not config.ssl_verify


@pytest.mark.usefixtures("disable_dot_env")
def test_conda_context_priority_config_toml(
    condarc_path: Path,
    config_toml: Path,
    monkeypatch: MonkeyPatch,
) -> None:
    condarc_path.write_text(
        dedent("""\
        ssl_verify: false
        proxy_servers:
            http: 127.0.0.1:80
            https: 127.0.0.1:80
    """)
    )
    conda.base.context.reset_context()

    config_toml.write_text(
        dedent("""\
        [plugin.auth]
        ssl_verify = true
    """)
    )

    # ~/.anaconda/config.toml overrides inherited settings
    default = AnacondaAuthConfig()
    assert default.ssl_verify
    assert default.proxy_servers == {"http": "127.0.0.1:80", "https": "127.0.0.1:80"}
    assert default.client_cert is None
    assert default.client_cert_key is None

    # Anaconda Auth secrets and env vars are higher priority
    monkeypatch.setenv("ANACONDA_AUTH_CLIENT_CERT", "/path/to/cert")
    default = AnacondaAuthConfig()
    assert default.ssl_verify
    assert default.proxy_servers == {"http": "127.0.0.1:80", "https": "127.0.0.1:80"}
    assert default.client_cert == "/path/to/cert"
    assert default.client_cert_key is None

    # Finally, init kwargs are highest priority
    yes_verified = AnacondaAuthConfig(ssl_verify=False, proxy_servers=None)
    assert not yes_verified.ssl_verify
    assert yes_verified.proxy_servers is None
    assert yes_verified.client_cert == "/path/to/cert"
    assert yes_verified.client_cert_key is None


@pytest.mark.usefixtures("disable_dot_env")
def test_conda_context_priority_sites(
    condarc_path: Path,
    config_toml: Path,
    monkeypatch: MonkeyPatch,
) -> None:
    condarc_path.write_text(
        dedent("""\
        ssl_verify: false
        proxy_servers:
            http: 127.0.0.1:80
            https: 127.0.0.1:80
    """)
    )
    conda.base.context.reset_context()

    config_toml.write_text(
        dedent("""\
        default_site = "anaconda.com"
        [sites."anaconda.com"]

        [sites.verified]
        ssl_verify = true

        [sites.proxied.proxy_servers]
        http = "proxy:80"
        https = "proxy:80"
    """)
    )

    # In the absence of an explicit setting inherit from conda context
    default = AnacondaAuthConfig()
    assert not default.ssl_verify
    assert default.proxy_servers == {"http": "127.0.0.1:80", "https": "127.0.0.1:80"}
    assert default.client_cert is None
    assert default.client_cert_key is None

    # ~/.anaconda/config.toml overrides inherited settings
    verified = AnacondaAuthConfig(site="verified")
    assert verified.ssl_verify
    assert verified.proxy_servers == {"http": "127.0.0.1:80", "https": "127.0.0.1:80"}
    assert verified.client_cert is None
    assert verified.client_cert_key is None

    proxied = AnacondaAuthConfig(site="proxied")
    assert not proxied.ssl_verify
    assert proxied.proxy_servers == {"http": "proxy:80", "https": "proxy:80"}
    assert proxied.client_cert is None
    assert proxied.client_cert_key is None

    # Anaconda Auth secrets and env vars are higher priority
    monkeypatch.setenv("ANACONDA_AUTH_SSL_VERIFY", "true")
    monkeypatch.setenv("ANACONDA_AUTH_CLIENT_CERT", "/path/to/cert")

    default = AnacondaAuthConfig()
    assert default.ssl_verify
    assert default.proxy_servers == {"http": "127.0.0.1:80", "https": "127.0.0.1:80"}
    assert default.client_cert == "/path/to/cert"
    assert default.client_cert_key is None

    verified = AnacondaAuthConfig(site="verified")
    assert verified.ssl_verify
    assert verified.proxy_servers == {"http": "127.0.0.1:80", "https": "127.0.0.1:80"}
    assert verified.client_cert == "/path/to/cert"
    assert verified.client_cert_key is None

    proxied = AnacondaAuthConfig(site="proxied")
    assert proxied.ssl_verify
    assert proxied.proxy_servers == {"http": "proxy:80", "https": "proxy:80"}
    assert proxied.client_cert == "/path/to/cert"
    assert proxied.client_cert_key is None

    # Finally, init kwargs are highest priority
    monkeypatch.delenv("ANACONDA_AUTH_SSL_VERIFY")
    yes_verified = AnacondaAuthConfig(ssl_verify=True, proxy_servers=None)
    assert yes_verified.ssl_verify
    assert yes_verified.proxy_servers is None
    assert yes_verified.client_cert == "/path/to/cert"
    assert yes_verified.client_cert_key is None



from __future__ import annotations

import os
import traceback
import warnings
from collections import defaultdict
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import IO
from typing import Any
from typing import Generator
from typing import Mapping
from typing import NamedTuple
from typing import Protocol
from typing import Sequence
from typing import cast
from uuid import UUID
from uuid import uuid4

import jwt
import pytest
import typer
from click.testing import Result
from dotenv import load_dotenv
from keyring.backend import KeyringBackend
from pytest import MonkeyPatch
from pytest_mock import MockerFixture
from requests_mock import Mocker as RequestMocker
from typer.testing import CliRunner

from anaconda_auth.client import BaseClient
from anaconda_auth.repo import OrganizationData
from anaconda_auth.repo import TokenCreateResponse
from anaconda_auth.repo import TokenInfoResponse
from anaconda_auth.token import TokenInfo
from anaconda_cli_base.cli import app

load_dotenv()


def is_conda_installed():
    try:
        import conda  # noqa: F401

        return True
    except ImportError:
        return False


class MockedKeyring(KeyringBackend):
    """A high-priority in-memory keyring backend for testing"""

    priority = 10000.0  # type: ignore
    _data: dict = defaultdict(dict)

    def __init__(self) -> None:
        super().__init__()

    def set_password(self, service: str, username: str, password: str) -> None:
        self._data[service][username] = password

    def get_password(self, service: str, username: str) -> str | None:
        password = self._data.get(service, {}).get(username, None)
        return password

    def delete_password(self, service: str, username: str) -> None:
        _ = self._data.get(service, {}).pop(username)


@pytest.fixture(autouse=True)
def clear_mocked_keyring() -> None:
    MockedKeyring._data = defaultdict(dict)


@pytest.fixture(autouse=True)
def set_keyring_name(mocker: MockerFixture) -> None:
    mocker.patch("anaconda_auth.token.KEYRING_NAME", "Anaconda Test")


@pytest.fixture
def outdated_api_key() -> str:
    # This is an old token from the dev system that will always be out-of-date
    api_key = (
        "eyJhbGciOiJSUzI1NiIsImtpZCI6ImRlZmF1bHQiLCJ0eXAiOiJKV1QifQ"
        ".eyJleHAiOjE2ODkwODg3ODYsInN1YiI6ImQwNGMzNTZiLWFmZDItNGIzZ"
        "S04MGYyLTQwMzExM2UwOTc0YiJ9.tTi_gttpQWhiTy_Uh0bDohN34mqd_6"
        "AHvyXf8_R5PFxjI-z9Ei0S3XCm9siP0RfyJx2j08SRs3FwXpkT8b8jP__C"
        "h-Y4K-zXYksZnTGcQ77YhKQCoKpGSpGlE4yD6gRXRRDT7bHs4H7gf4e6iD"
        "1Vdcq0yx5-5h-CbBgSwS9LSpJ_HDZBUy-xbRrw0aD36aQ5qs6huswgCOQa"
        "9YrYfsrSbZW8uY48LAt5Y69t8x1twNBI5_Cumx-JEZuDLQxq7HQp7wKldE"
        "tbycV5uemKjyR1Qeuva2zCKYB3FEXdTEiWHhTzhSQ-3-xjUrIZvpfGJd3G"
        "CzXlkUhpeDoj2KbSN-Lq0Q"
    )
    return api_key


@pytest.fixture
def outdated_token_info(outdated_api_key: str) -> TokenInfo:
    return TokenInfo(api_key=outdated_api_key, domain="mocked-domain")


@pytest.fixture()
def tmp_cwd(monkeypatch: MonkeyPatch, tmp_path: Path) -> Path:
    """Create & return a temporary directory after setting current working directory to it."""
    monkeypatch.chdir(tmp_path)
    return tmp_path


@pytest.fixture(scope="session")
def is_not_none() -> Any:
    """
    An object that can be used to test whether another is None.

    This is particularly useful when testing contents of collections, e.g.:

    ```python
    def test_data(data, is_not_none):
        assert data == {"some_key": is_not_none, "some_other_key": 5}
    ```

    """

    class _NotNone:
        def __eq__(self, other: Any) -> bool:
            return other is not None

    return _NotNone()


@pytest.fixture
def disable_dot_env(mocker: MockerFixture) -> None:
    from anaconda_cli_base.config import AnacondaBaseSettings

    mocker.patch.dict(AnacondaBaseSettings.model_config, {"env_file": ""})


@pytest.fixture(autouse=True)
def disable_config_toml(tmp_path: Path, monkeypatch: MonkeyPatch):
    monkeypatch.setenv("ANACONDA_CONFIG_TOML", str(tmp_path / "empty-config.toml"))


@pytest.fixture
def api_key() -> str | None:
    return os.getenv("TEST_API_KEY")


@pytest.fixture()
def integration_test_client(api_key: str | None) -> BaseClient:
    c = BaseClient(api_key=api_key)
    return c


@pytest.fixture()
def save_api_key_to_token(api_key: str | None) -> Generator[None, None, None]:
    from anaconda_auth.config import AnacondaAuthConfig

    conf = AnacondaAuthConfig()
    token = TokenInfo(api_key=api_key, domain=conf.domain)
    token.save()
    yield


def pytest_addoption(parser):  # type: ignore
    """Defines custom CLI options."""
    parser.addoption(
        "--integration",
        action="store_true",
        dest="integration",
        default=False,
        help="enable integration tests",
    )
    parser.addoption(
        "--error-on-pending-deprecations",
        action="store_true",
        default=False,
        help="Treat PendingDeprecationWarnings from anaconda_auth as errors",
    )


def pytest_configure(config):
    """Customize warning filters to ignore PendingDeprecationWarnings that are
    raised by other conda plugins, while allowing us to still catch our own.
    """
    if config.getoption("--error-on-pending-deprecations"):
        warnings.simplefilter("error", PendingDeprecationWarning)

        original_warn = warnings.warn

        def custom_warn(message, category=UserWarning, stacklevel=1, source=None):
            if isinstance(category, type) and issubclass(
                category, PendingDeprecationWarning
            ):
                stack = traceback.extract_stack()
                # Ignore the warning if any of the known plugins matches
                for frame in stack:
                    if "conda_libmamba_solver" in frame.filename:
                        return
                    if "anaconda_anon_usage" in frame.filename:
                        return

            original_warn(message, category, stacklevel + 1, source)

        warnings.warn = custom_warn


def pytest_collection_modifyitems(config, items):  # type: ignore
    """Auto-mark each test in the integration directory, and enable them based on CLI flag."""
    integration_test_root_dir = Path(__file__).parent / "integration"
    run_integration_tests = config.getoption("--integration")
    for item in items:
        # Here, we add a marker to any test in the "tests/integration" directory
        if integration_test_root_dir in Path(item.fspath).parents:
            item.add_marker(pytest.mark.integration)

        # Add a skip marker if the CLI option is not used. We use an additional marker so that we can
        # independently select integrations with `pytest -m integration` and enable them with `--integration`.
        if "integration" in item.keywords and not run_integration_tests:
            item.add_marker(pytest.mark.skip(reason="need --integration to run"))


@pytest.fixture
def with_aau_token(mocker: MockerFixture) -> None:
    mocker.patch("anaconda_auth.config.AnacondaAuthSite.aau_token", "anon-token")


@pytest.fixture
def without_aau_token(mocker: MockerFixture) -> None:
    mocker.patch("anaconda_auth.config.AnacondaAuthSite.aau_token", None)


class MockResponse:
    def __init__(
        self,
        *,
        status_code: int,
        json_data: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ):
        self.status_code = status_code
        self.json_data = json_data
        self.headers = headers or {}

    def json(self) -> dict[str, Any]:
        return self.json_data or {}


class MockedRequest:
    def __init__(
        self,
        *,
        response_status_code: int,
        response_data: dict[str, Any] | None = None,
        response_headers: dict[str, str] | None = None,
    ) -> None:
        self.response_status_code = response_status_code
        self.response_data = response_data
        self.response_headers = response_headers
        self.called_with_args: tuple[Any] = ()  # type: ignore
        self.called_with_kwargs: dict[str, Any] = {}

    def __call__(self, *args: Any, **kwargs: Any) -> MockResponse:
        self.called_with_args = args  # type: ignore
        self.called_with_kwargs = kwargs
        return MockResponse(
            status_code=self.response_status_code,
            json_data=self.response_data,
            headers=self.response_headers,
        )


class CLIInvoker(Protocol):
    def __call__(
        self,
        # app: typer.Typer,
        args: str | Sequence[str] | None = None,
        input: bytes | str | IO[Any] | None = None,
        env: Mapping[str, str] | None = None,
        catch_exceptions: bool = True,
        color: bool = False,
        **extra: Any,
    ) -> Result: ...


@pytest.fixture()
def invoke_cli(tmp_cwd: Path) -> CLIInvoker:
    """Returns a function, which can be used to call the CLI from within a temporary directory."""

    runner = CliRunner()

    return partial(runner.invoke, cast(typer.Typer, app))


@pytest.fixture
def config_toml(
    tmp_path: Path, monkeypatch: MonkeyPatch
) -> Generator[Path, None, None]:
    config_file = tmp_path / "config.toml"
    monkeypatch.setenv("ANACONDA_CONFIG_TOML", str(config_file))
    yield config_file


class CondaRCPaths(NamedTuple):
    user: Path
    prefix: Path
    sites: Path


@pytest.fixture()
def conda_search_path(monkeypatch, tmp_path):
    """
    Build an independent, empty set of conda configuration locations
    that we can write to to emulate various combinations of user,
    prefix, and site configuration data, while remaining isolated
    from the local user's existing configuration.
    """
    user_path = tmp_path / ".condarc"
    prefix_path = tmp_path / "prefix" / "condarc.d"
    sites_path = tmp_path / ".conda" / "condarc.d"
    user_path.touch()
    prefix_path.mkdir(parents=True)
    sites_path.mkdir(parents=True)

    # Patch where the condarc object looks for the user file
    from anaconda_auth._conda import condarc as condarc_module

    monkeypatch.setattr(condarc_module, "DEFAULT_CONDARC_PATH", user_path)

    # Patch where the default channel_settings config is written
    from anaconda_auth._conda import config as plugin_config

    config_path = prefix_path / "anaconda-auth.yml"
    monkeypatch.setattr(plugin_config, "PREFIX_CONDARC_PATH", config_path)

    # Patch the handling of conda CLI arguments to pass the path to the condarc file
    from anaconda_auth._conda import repo_config

    orig_get_condarc_args = repo_config._get_condarc_args

    def _new_get_condarc_args(*args, **kwargs) -> None:
        return orig_get_condarc_args(condarc_file=str(user_path))

    monkeypatch.setattr(repo_config, "_get_condarc_args", _new_get_condarc_args)

    if is_conda_installed():
        # Patch the default conda search path
        from conda.base import context

        # In the standard conda search path, CONDA_ROOT/condarc.d has
        search_path = [prefix_path, sites_path, user_path, prefix_path]
        monkeypatch.setattr(context, "SEARCH_PATH", search_path)

        # Patch reset_context function such that it only loads config from our temp file
        orig_reset_context = context.reset_context

        def _new_reset_context(*args, **kwargs):
            return orig_reset_context(search_path)

        monkeypatch.setattr(context, "reset_context", _new_reset_context)

        # Reset the context object with these new settings
        context.reset_context()

    yield CondaRCPaths(user_path, prefix_path, sites_path)


@pytest.fixture()
def condarc_path(conda_search_path):
    """
    This is the old condarc_path fixture used in a lot of tests that need
    nothing more than the user condarc pathname.
    """
    yield conda_search_path.user


@pytest.fixture()
def org_name() -> str:
    return "test-org-name"


@pytest.fixture()
def token_is_installed(org_name: str, valid_api_key: TokenInfo) -> TokenInfo:
    valid_api_key.set_repo_token(org_name=org_name, token="test-token")
    valid_api_key.save()
    return valid_api_key


@pytest.fixture(params=[True, False])
def no_tokens_installed(
    request, mocker: MockerFixture, valid_api_key: TokenInfo
) -> None:
    if request.param:
        # Remove the API key
        valid_api_key.delete()
    else:
        # Models the situation where we have a valid API key but it has no attached repo tokens.
        pass

    # No legacy tokens either
    mocker.patch(
        "anaconda_auth._conda.repo_config.read_binstar_tokens",
        return_value={},
    )


@pytest.fixture()
def token_does_not_exist_in_service(
    requests_mock: RequestMocker, org_name: str
) -> None:
    requests_mock.get(
        f"https://anaconda.com/api/organizations/{org_name}/ce/current-token",
        status_code=404,
    )


@pytest.fixture()
def token_exists_in_service(
    requests_mock: RequestMocker, org_name: str
) -> TokenInfoResponse:
    token_info = TokenInfoResponse(
        id=uuid4(), expires_at=datetime(year=2025, month=1, day=1)
    )
    requests_mock.get(
        f"https://anaconda.com/api/organizations/{org_name}/ce/current-token",
        json=token_info.model_dump(mode="json"),
    )
    return token_info


@pytest.fixture()
def token_created_in_service(
    requests_mock: RequestMocker, org_name: str
) -> TokenCreateResponse:
    test_token = "test-token"
    payload = {"token": test_token, "expires_at": "2025-01-01T00:00:00"}
    requests_mock.put(
        f"https://anaconda.com/api/organizations/{org_name}/ce/current-token",
        json=payload,
    )
    return TokenCreateResponse(**payload)


@pytest.fixture()
def user_has_one_org(
    requests_mock: RequestMocker, org_name: str, business_org_id: UUID
) -> TokenCreateResponse:
    requests_mock.get(
        "https://anaconda.com/api/organizations/my",
        json=[
            {
                "id": str(business_org_id),
                "name": org_name,
                "title": "My Cool Organization",
            }
        ],
    )
    return [
        OrganizationData(
            id=business_org_id, name=org_name, title="My Cool Organization"
        )
    ]


@pytest.fixture()
def user_has_multiple_orgs(
    requests_mock: RequestMocker, org_name: str, business_org_id: UUID
) -> TokenCreateResponse:
    first_id = uuid4()
    requests_mock.get(
        "https://anaconda.com/api/organizations/my",
        json=[
            {
                "id": str(first_id),
                "name": "first-org",
                "title": "My First Organization",
            },
            {
                "id": str(business_org_id),
                "name": org_name,
                "title": "My Business Organization",
            },
        ],
    )
    return [
        OrganizationData(id=first_id, name="first-org", title="My First Organizatoin"),
        OrganizationData(
            id=business_org_id, name=org_name, title="My Business Organization"
        ),
    ]


@pytest.fixture()
def user_has_no_orgs(
    requests_mock: RequestMocker, user_has_no_subscriptions: None
) -> list[OrganizationData]:
    requests_mock.get(
        "https://anaconda.com/api/organizations/my",
        json=[],
    )
    return []


@pytest.fixture()
def business_org_id() -> UUID:
    return uuid4()


@pytest.fixture()
def user_has_starter_subscription(
    request, requests_mock: RequestMocker, business_org_id: UUID
) -> None:
    requests_mock.get(
        "https://anaconda.com/api/account",
        json={
            "subscriptions": [
                {
                    "org_id": str(business_org_id),
                    "product_code": "starter_subscription",
                }
            ]
        },
    )


@pytest.fixture()
def user_has_no_subscriptions(requests_mock: RequestMocker) -> None:
    requests_mock.get("https://anaconda.com/api/account", json={})


@pytest.fixture()
def valid_api_key():
    token_info = TokenInfo.load(create=True)
    exp = int(datetime(2099, 1, 1).timestamp())
    token_info.api_key = jwt.encode({"exp": exp}, key="secret", algorithm="HS256")
    token_info.save()
    return token_info



"""
Device Code Flow implementation for OAuth 2.0 device authorization grant (RFC 8628).
"""

import json
import time
from typing import Dict
from typing import Optional

import requests
from pydantic import BaseModel

from anaconda_auth.client import BaseClient
from anaconda_auth.config import AnacondaAuthSite
from anaconda_auth.exceptions import DeviceFlowDenied
from anaconda_auth.exceptions import DeviceFlowError
from anaconda_auth.exceptions import DeviceFlowTimeout


class DeviceAuthorizationResponse(BaseModel):
    device_code: str
    user_code: str
    verification_uri: str
    verification_uri_complete: str
    expires_in: int = 60
    interval: int = 5


class DeviceCodeFlow:
    """
    OAuth 2.0 Device Code Flow implementation.

    This implements RFC 8628 for devices that are either browserless
    or have limited input capabilities.
    """

    config: AnacondaAuthSite
    authorize_response: Optional[DeviceAuthorizationResponse]

    def __init__(self, config: AnacondaAuthSite):
        """
        Initialize device code flow.

        Args:
            config: Configuration for the client
        """
        self.config = config
        self.client = BaseClient(site=config)

        # Device authorization response data
        self.authorize_response = None

    def initiate_device_authorization(self) -> DeviceAuthorizationResponse:
        """
        Initiate device authorization request.

        Returns:
            Tuple of (user_code, verification_uri) to display to user
        """

        data = {"client_id": self.config.client_id, "scope": "openid"}

        if self.config.oidc.device_authorization_endpoint is None:
            raise DeviceFlowError("Server does not support device authorization")

        try:
            response = self.client.post(
                self.config.oidc.device_authorization_endpoint,
                data=data,
                verify=self.config.ssl_verify,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
            )
            response.raise_for_status()

            auth_response = response.json()

            self.authorize_response = DeviceAuthorizationResponse(**auth_response)
            return self.authorize_response

        except requests.RequestException as e:
            raise DeviceFlowError(f"Device authorization request failed: {e}")
        except KeyError as e:
            raise DeviceFlowError(f"Missing required field in response: {e}")

    def poll_for_token(self) -> Dict[str, str]:
        """
        Poll the token endpoint until authorization is complete.

        Returns:
            Token response containing access_token, etc.
        """
        if not self.authorize_response:
            raise DeviceFlowError("Must call initiate_device_authorization first")

        start_time = time.time()
        expires_in = self.authorize_response.expires_in
        interval = self.authorize_response.interval

        while time.time() - start_time < expires_in:
            try:
                token_response = self._request_token()
                return token_response

            except DeviceFlowTimeout:
                raise
            except DeviceFlowDenied:
                raise
            except DeviceFlowError as e:
                # Check for authorization_pending
                if "authorization_pending" in str(e).lower():
                    time.sleep(interval)
                    continue
                elif "slow_down" in str(e).lower():
                    # Server asked us to slow down
                    interval = min(interval + 5, 30)
                    time.sleep(interval)
                    continue
                else:
                    raise

        raise DeviceFlowTimeout("Device authorization timed out")

    def _request_token(self) -> Dict[str, str]:
        """Make a single token request."""
        if not self.authorize_response:
            raise DeviceFlowError("Must call initiate_device_authorization first")

        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": self.authorize_response.device_code,
            "client_id": self.config.client_id,
        }

        response = self.client.post(
            self.config.oidc.token_endpoint,
            data=data,
            verify=self.config.ssl_verify,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )

        if response.status_code == 200:
            return response.json()

        # Handle error responses
        try:
            error_data = response.json()
            error_code = error_data.get("error", "unknown_error")
            error_description = error_data.get("error_description", "")

            if error_code == "authorization_pending":
                raise DeviceFlowError("authorization_pending")
            elif error_code == "slow_down":
                raise DeviceFlowError("slow_down")
            elif error_code == "expired_token":
                raise DeviceFlowTimeout("Device code expired")
            elif error_code == "access_denied":
                raise DeviceFlowDenied("User denied authorization")
            else:
                raise DeviceFlowError(
                    f"Token request failed: {error_code} - {error_description}"
                )

        except json.JSONDecodeError:
            response.raise_for_status()

        raise DeviceFlowError("Token request failed")



from pathlib import Path
from textwrap import dedent

import pytest
from pytest import MonkeyPatch
from pytest_mock import MockerFixture
from requests_mock import Mocker as RequestMocker

from anaconda_auth.client import BaseClient
from anaconda_auth.config import AnacondaAuthConfig
from anaconda_auth.config import AnacondaAuthSite
from anaconda_auth.config import AnacondaAuthSitesConfig
from anaconda_auth.config import AnacondaCloudConfig
from anaconda_auth.config import Sites
from anaconda_auth.exceptions import UnknownSiteName


@pytest.fixture(
    autouse=True,
    params=[
        "with-device-authorization-endpoint",
        "without-device-authorization-endpoint",
    ],
)
def mock_openid_configuration(request, requests_mock: RequestMocker):
    """Mock return value of openid configuration to prevent requiring actual network calls."""
    config = AnacondaAuthConfig()
    expected = {
        "authorization_endpoint": f"https://auth.{config.domain}/api/auth/oauth2/authorize",
        "token_endpoint": f"https://auth.{config.domain}/api/auth/oauth2/token",
    }
    # This field was added to the openid configuration to support device auth, but is
    # not present on anaconda.org, so we need to test it as optional. Remove once
    # we don't need to special case this.
    if request.param == "with-device-authorization-endpoint":
        expected["device_authorization_endpoint"] = (
            f"https://auth.{config.domain}/api/auth/oauth2/device/authorize"
        )
    requests_mock.get(url=config.well_known_url, json=expected)


def test_well_known_headers(mocker: MockerFixture) -> None:
    spy = mocker.spy(BaseClient, "get")

    config = AnacondaAuthConfig()
    assert config.oidc
    spy.assert_called_once()
    assert spy.spy_return.request.headers.get("User-Agent").startswith("anaconda-auth")
    assert not spy.call_args.kwargs["auth"]


@pytest.mark.parametrize("prefix", ["ANACONDA_AUTH", "ANACONDA_CLOUD"])
def test_docker_secret_over_default(
    tmp_path: Path, monkeypatch: MonkeyPatch, prefix: str
) -> None:
    monkeypatch.setitem(AnacondaAuthConfig.model_config, "secrets_dir", tmp_path)
    monkeypatch.setitem(AnacondaCloudConfig.model_config, "secrets_dir", tmp_path)
    key = f"{prefix}_API_KEY"
    with open(tmp_path / key.lower(), "w") as fp:
        fp.write("set-in-docker-secret")
    config = AnacondaAuthConfig()
    assert config.api_key == "set-in-docker-secret"


@pytest.mark.parametrize("prefix", ["ANACONDA_AUTH", "ANACONDA_CLOUD"])
def test_docker_secret_no_match(
    tmp_path: Path, monkeypatch: MonkeyPatch, prefix: str
) -> None:
    monkeypatch.setitem(AnacondaAuthConfig.model_config, "secrets_dir", tmp_path)
    monkeypatch.setitem(AnacondaCloudConfig.model_config, "secrets_dir", tmp_path)
    key = f"{prefix}_NONEXISTENT"
    with open(tmp_path / key.lower(), "w") as fp:
        fp.write("set-in-docker-secret")
    config = AnacondaAuthConfig()
    assert not hasattr(config, "nonexistent")


@pytest.mark.parametrize("prefix", ["ANACONDA_AUTH", "ANACONDA_CLOUD"])
def test_env_variable_over_default(monkeypatch: MonkeyPatch, prefix: str) -> None:
    monkeypatch.setenv(f"{prefix}_DOMAIN", "set-in-env")
    config = AnacondaAuthConfig()
    assert config.domain == "set-in-env"


@pytest.mark.parametrize("prefix", ["ANACONDA_AUTH", "ANACONDA_CLOUD"])
def test_env_variable_over_docker(
    tmp_path: Path, monkeypatch: MonkeyPatch, prefix: str
) -> None:
    monkeypatch.setitem(AnacondaAuthConfig.model_config, "secrets_dir", tmp_path)
    monkeypatch.setitem(AnacondaCloudConfig.model_config, "secrets_dir", tmp_path)
    key = f"{prefix}_API_KEY"
    monkeypatch.setenv(key, "set-in-env")
    with open(tmp_path / key.lower(), "w") as fp:
        fp.write("set-in-docker-secret")
    config = AnacondaAuthConfig()
    assert config.api_key == "set-in-env"


@pytest.mark.parametrize("prefix", ["ANACONDA_AUTH", "ANACONDA_CLOUD"])
def test_init_arg_over_all(
    tmp_path: Path, monkeypatch: MonkeyPatch, prefix: str
) -> None:
    monkeypatch.setitem(AnacondaAuthConfig.model_config, "secrets_dir", tmp_path)
    monkeypatch.setitem(AnacondaCloudConfig.model_config, "secrets_dir", tmp_path)
    key = f"{prefix}_DOMAIN"
    monkeypatch.setenv(key, "set-in-env")
    with open(tmp_path / key.lower(), "w") as fp:
        fp.write("set-in-docker-secret")
    config = AnacondaAuthConfig(domain="set-in-init")
    assert config.domain == "set-in-init"


def test_auth_domain_default_behavior() -> None:
    config = AnacondaAuthConfig()
    assert config.domain == config.auth_domain


def test_override_auth_domain_env_variable(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv(
        "ANACONDA_AUTH_AUTH_DOMAIN_OVERRIDE", "another-auth.anaconda.com"
    )
    config = AnacondaAuthConfig()
    assert config.auth_domain == "another-auth.anaconda.com"


@pytest.mark.usefixtures("disable_dot_env", "config_toml")
def test_default_site_no_config() -> None:
    config = AnacondaAuthSitesConfig()

    assert config.sites == Sites({"anaconda.com": AnacondaAuthSite()})
    assert config.default_site == "anaconda.com"
    # The classes are not literally the same but the values are
    assert (
        AnacondaAuthSitesConfig.load_site().model_dump()
        == AnacondaAuthConfig().model_dump()
    )


@pytest.mark.usefixtures("disable_dot_env", "config_toml")
def test_unknown_site() -> None:
    config = AnacondaAuthSitesConfig()

    with pytest.raises(UnknownSiteName):
        _ = config.sites["unknown-site"]


@pytest.mark.usefixtures("disable_dot_env")
def test_default_site_with_plugin_config(config_toml: Path) -> None:
    config_toml.write_text(
        dedent(
            """\
            [plugin.auth]
            domain = "localhost"
            ssl_verify = false
            """
        )
    )
    config = AnacondaAuthSitesConfig()

    assert config.sites == Sites({"anaconda.com": AnacondaAuthSite()})
    assert config.default_site == "anaconda.com"

    default_site = AnacondaAuthSitesConfig.load_site()
    # this default site is identical to the AnacondaAuthConfig()
    assert default_site.model_dump() == AnacondaAuthConfig().model_dump()
    assert default_site.domain == "localhost"
    assert not default_site.ssl_verify


@pytest.mark.usefixtures("disable_dot_env")
def test_extra_site_config(config_toml: Path) -> None:
    config_toml.write_text(
        dedent(
            """\
            [sites.local]
            domain = "localhost"
            ssl_verify = false
            """
        )
    )

    config = AnacondaAuthSitesConfig()

    local = AnacondaAuthSite(
        site="local",
        domain="localhost",
        ssl_verify=False,
    )

    assert config.sites == Sites({"local": local})

    assert config.default_site == "local"
    # This default site anaconda.com is identical to the AnacondaAuthConfig()
    assert (
        AnacondaAuthSitesConfig.load_site().model_dump()
        == AnacondaAuthConfig().model_dump()
    )

    site = AnacondaAuthSitesConfig.load_site(site="local")
    assert site.model_dump() == local.model_dump()
    assert site.domain == "localhost"
    assert AnacondaAuthSitesConfig.load_site(
        site="local"
    ) == AnacondaAuthSitesConfig.load_site(site="localhost")


@pytest.mark.usefixtures("disable_dot_env")
def test_default_extra_site_config(config_toml: Path) -> None:
    config_toml.write_text(
        dedent(
            """\
            default_site = "local"

            [sites.local]
            domain = "localhost"
            auth_domain_override = "auth-local"
            ssl_verify = false
            """
        )
    )

    config = AnacondaAuthSitesConfig()

    local = AnacondaAuthSite(
        site="local",
        domain="localhost",
        ssl_verify=False,
        auth_domain_override="auth-local",
    )

    assert config.sites == Sites({"local": local})

    assert config.sites["local"].model_dump() == local.model_dump()
    assert config.default_site == "local"
    assert AnacondaAuthSitesConfig.load_site().model_dump() == local.model_dump()


@pytest.mark.usefixtures("disable_dot_env")
def test_duplicate_domain_lookup_fail(config_toml: Path) -> None:
    config_toml.write_text(
        dedent(
            """\
            [sites.local1]
            domain = "localhost"
            ssl_verify = false

            [sites.local2]
            domain = "localhost"
            ssl_verify = true
            """
        )
    )

    config = AnacondaAuthSitesConfig()

    assert config.sites["local1"].ssl_verify is False
    assert config.sites["local2"].ssl_verify is True

    with pytest.raises(ValueError):
        _ = config.sites["localhost"]


@pytest.mark.usefixtures("disable_dot_env")
def test_site_inherits_from_plugin_auth_config(config_toml: Path) -> None:
    config_toml.write_text(
        dedent(
            """\
            [plugin.auth]
            domain = "foo.com"
            client_id = "foo"
            ssl_verify = false

            [sites.local]
            domain = "localhost"
            auth_domain_override = "auth-local"
            """
        )
    )

    config = AnacondaAuthSitesConfig()

    local = config.sites["local"]
    assert local.domain == "localhost"
    # [plugin.auth] provide backfills for unset values in all [sites.<name>]
    assert local.client_id == "foo"
    assert not local.ssl_verify


@pytest.mark.usefixtures("disable_dot_env")
def test_override_site_with_auth_env_vars(
    config_toml: Path, monkeypatch: MonkeyPatch
) -> None:
    config_toml.write_text(
        dedent(
            """\
            [plugin.auth]
            domain = "foo.com"
            client_id = "baz"
            ssl_verify = false

            [sites.local]
            domain = "localhost"
            auth_domain_override = "auth-local"
            client_id = "bar"
            """
        )
    )

    monkeypatch.setenv("ANACONDA_AUTH_API_KEY", "foo")

    config = AnacondaAuthSitesConfig()

    assert config.sites["local"].api_key == "foo"
    assert config.sites["local"].domain == "localhost"
    assert config.sites["local"].auth_domain_override == "auth-local"
    assert config.sites["local"].client_id == "bar"
    assert not config.sites["local"].ssl_verify

    # Finally, ANACONDA_AUTH_* env vars even override the site configuration
    monkeypatch.setenv("ANACONDA_AUTH_CLIENT_ID", "override-in-env")

    config = AnacondaAuthSitesConfig()
    local = config.sites["local"]
    assert config.sites["local"].domain == "localhost"
    assert config.sites["local"].auth_domain_override == "auth-local"
    assert local.client_id == "override-in-env"
    assert not local.ssl_verify



class AuthenticationError(Exception):
    pass


class InvalidTokenError(AuthenticationError):
    pass


class TokenNotFoundError(Exception):
    pass


class LoginRequiredError(Exception):
    pass


class TokenExpiredError(Exception):
    pass


class UnknownSiteName(Exception):
    pass


class DeviceFlowError(Exception):
    """Base exception for device flow errors."""

    pass


class DeviceFlowTimeout(DeviceFlowError):
    """Device authorization timed out."""

    pass


class DeviceFlowDenied(DeviceFlowError):
    """User denied device authorization."""

    pass



from typing import TYPE_CHECKING
from typing import Any
from typing import Optional

from requests.adapters import HTTPAdapter as BaseHttpAdapter

if TYPE_CHECKING:
    from ssl import SSLContext

    from urllib3 import PoolManager


class _SSLContextAdapterMixin:
    """Mixin to add the ``ssl_context`` constructor argument to HTTP adapters.

    The additional argument is forwarded directly to the pool manager. This allows us
    to dynamically decide what SSL store to use at runtime, which is used to implement
    the optional ``truststore`` backend.
    """

    def __init__(
        self,
        *,
        ssl_context: Optional["SSLContext"] = None,
        **kwargs: Any,
    ) -> None:
        self._ssl_context = ssl_context
        super().__init__(**kwargs)

    def init_poolmanager(
        self,
        connections: int,
        maxsize: int,
        block: bool = False,
        **pool_kwargs: Any,
    ) -> "PoolManager":
        if self._ssl_context is not None:
            pool_kwargs.setdefault("ssl_context", self._ssl_context)
        return super().init_poolmanager(  # type: ignore[misc]
            connections=connections,
            maxsize=maxsize,
            block=block,
            **pool_kwargs,
        )


class HTTPAdapter(_SSLContextAdapterMixin, BaseHttpAdapter):
    pass



from datetime import datetime
from uuid import UUID

import pytest
from pytest_mock import MockerFixture
from requests_mock import Mocker as RequestMocker

pytest.importorskip("conda")

# ruff: noqa: E402
from anaconda_auth.repo import OrganizationData
from anaconda_auth.repo import RepoAPIClient
from anaconda_auth.repo import TokenCreateResponse
from anaconda_auth.repo import TokenInfoResponse


@pytest.fixture(autouse=True)
def mock_do_auth_flow(mocker: MockerFixture) -> None:
    mocker.patch(
        "anaconda_auth.actions._do_auth_flow",
        return_value="test-access-token",
    )


@pytest.fixture(
    autouse=True, params=["security_subscription", "commercial_subscription"]
)
def user_has_business_subscription(
    request, requests_mock: RequestMocker, org_name: str, business_org_id: UUID
) -> None:
    requests_mock.get(
        "https://anaconda.com/api/account",
        json={
            "subscriptions": [
                {
                    "org_id": str(business_org_id),
                    "product_code": request.param,
                }
            ]
        },
    )


@pytest.fixture(autouse=True)
def repodata_json_available_with_token(
    requests_mock: RequestMocker, token_created_in_service: TokenCreateResponse
) -> None:
    requests_mock.head(
        f"https://repo.anaconda.cloud/t/{token_created_in_service.token}/repo/main/noarch/repodata.json",
        status_code=200,
    )


def test_get_repo_token_info_no_token(
    org_name: str, token_does_not_exist_in_service: None
) -> None:
    client = RepoAPIClient()
    token_info = client._get_repo_token_info(org_name=org_name)
    assert token_info is None


def test_get_repo_token_info_has_token(
    org_name: str,
    token_exists_in_service: TokenInfoResponse,
) -> None:
    client = RepoAPIClient()
    token_info = client._get_repo_token_info(org_name=org_name)
    assert token_info == token_exists_in_service


def test_create_repo_token_info_has_token(
    org_name: str,
    token_created_in_service: TokenCreateResponse,
) -> None:
    client = RepoAPIClient()
    token_info = client._create_repo_token(org_name=org_name)
    assert token_info == token_created_in_service


def test_get_organizations_for_user(user_has_one_org: list[OrganizationData]) -> None:
    client = RepoAPIClient()
    organizations = client.get_organizations_for_user()
    assert organizations == user_has_one_org


def test_get_business_organizations_for_user(
    org_name: str,
    business_org_id: UUID,
    user_has_multiple_orgs: list[OrganizationData],
    user_has_business_subscription: None,
) -> None:
    client = RepoAPIClient()
    organizations = client.get_business_organizations_for_user()
    assert organizations == [
        OrganizationData(
            id=business_org_id, name=org_name, title="My Business Organization"
        )
    ]


def test_get_business_organizations_for_user_only_starter(
    org_name: str,
    business_org_id: UUID,
    user_has_multiple_orgs: list[OrganizationData],
    user_has_starter_subscription: None,
) -> None:
    client = RepoAPIClient()
    organizations = client.get_business_organizations_for_user()
    assert organizations == []


def test_issue_new_token_prints_success_message_via_client(
    org_name: str, mocker: MockerFixture, capsys: pytest.CaptureFixture
) -> None:
    client = RepoAPIClient()
    mocker.patch.object(client, "_get_repo_token_info", return_value=None)

    mock_response = mocker.MagicMock()
    mock_response.expires_at = datetime(2025, 12, 31)
    mocker.patch.object(client, "_create_repo_token", return_value=mock_response)
    client.issue_new_token(org_name=org_name)
    res = capsys.readouterr()
    expected_msg = "Your conda token has been installed and expires 2025-12-31 00:00:00. To view your token(s), you can use anaconda token list\n"

    assert expected_msg in res.out



import os
import sys
import warnings
from textwrap import dedent
from typing import List
from typing import Optional

import typer
from requests.exceptions import HTTPError
from requests.exceptions import JSONDecodeError
from rich.prompt import Confirm
from rich.syntax import Syntax

from anaconda_auth import __version__
from anaconda_auth.actions import login
from anaconda_auth.actions import logout
from anaconda_auth.client import BaseClient
from anaconda_auth.exceptions import TokenExpiredError
from anaconda_auth.token import TokenInfo
from anaconda_auth.token import TokenNotFoundError
from anaconda_cli_base.config import anaconda_config_path
from anaconda_cli_base.console import console
from anaconda_cli_base.exceptions import register_error_handler


def _continue_with_login() -> int:
    if sys.stdout.isatty():
        do_login = Confirm.ask("Continue with interactive login?", choices=["y", "n"])
        if do_login:
            login()
            return -1
        else:
            console.print(
                dedent("""
                To configure your credentials you can run
                  [green]anaconda login --at anaconda.com[/green]

                or set your API key using the [green]ANACONDA_AUTH_API_KEY[/green] env var

                or set
                """)
            )
            console.print(
                Syntax(
                    dedent(
                        """\
                        [plugin.auth]
                        api_key = "<api-key>"
                        """
                    ),
                    "toml",
                    background_color=None,
                )
            )
            console.print(f"in {anaconda_config_path()}")
    return 1


def _login_required_message(error_classifier: str) -> None:
    console.print(
        f"[bold][red]{error_classifier}[/red][/bold]: Login is required to complete this action."
    )


@register_error_handler(TokenNotFoundError)
def login_required(e: Exception) -> int:
    _login_required_message(e.__class__.__name__)
    return _continue_with_login()


@register_error_handler(TokenExpiredError)
def token_expired(e: Exception) -> int:
    console.print(
        f"[bold][red]{e.__class__.__name__}[/red][/bold]: Your login token has expired"
    )

    return _continue_with_login()


@register_error_handler(HTTPError)
def http_error(e: HTTPError) -> int:
    try:
        error_code = e.response.json().get("error", {}).get("code", "")
    except JSONDecodeError:
        error_code = ""

    if error_code == "auth_required":
        if "Authorization" in e.request.headers:
            console.print(
                "[bold][red]InvalidAuthentication:[/red][/bold] Your provided API Key or login token is invalid"
            )
        else:
            _login_required_message("AuthenticationMissingError")
        return _continue_with_login()
    else:
        console.print(f"[bold][red]{e.__class__.__name__}:[/red][/bold] {e}")
        return 1


def _override_default_site(at: Optional[str] = None) -> None:
    if at:
        os.environ["ANACONDA_DEFAULT_SITE"] = at


app = typer.Typer(
    name="auth",
    add_completion=False,
    help="Manage your Anaconda authentication",
    context_settings={
        "allow_extra_args": True,
        "ignore_unknown_options": True,
        "help_option_names": ["--help", "-h"],
    },
)


@app.callback(
    invoke_without_command=True,
    no_args_is_help=False,
)
def main(
    ctx: typer.Context,
    version: bool = typer.Option(False, "-V", "--version"),
    name: Optional[str] = typer.Option(
        None,
        "-n",
        "--name",
        hidden=True,
    ),
    organization: Optional[str] = typer.Option(
        None,
        "-o",
        "--org",
        "--organization",
        hidden=True,
    ),
    strength: Optional[str] = typer.Option(
        None,
        "--strength",
        hidden=True,
    ),
    strong: Optional[bool] = typer.Option(
        None,
        "--strong",
        hidden=True,
    ),
    weak: Optional[bool] = typer.Option(
        None,
        "-w",
        "--weak",
        hidden=True,
    ),
    url: Optional[str] = typer.Option(
        None,
        "--url",
        hidden=True,
    ),
    max_age: Optional[str] = typer.Option(
        None,
        "--max-age",
        hidden=True,
    ),
    scopes: Optional[str] = typer.Option(
        None,
        "-s",
        "--scopes",
        hidden=True,
    ),
    out: Optional[str] = typer.Option(
        None,
        "--out",
        hidden=True,
    ),
    list_scopes: Optional[bool] = typer.Option(
        None,
        "-x",
        "--list-scopes",
        hidden=True,
    ),
    list_tokens: Optional[bool] = typer.Option(
        None,
        "-l",
        "--list",
        hidden=True,
    ),
    remove: Optional[str] = typer.Option(
        None,
        "-r",
        "--remove",
        hidden=True,
    ),
    create: Optional[bool] = typer.Option(
        None,
        "-c",
        "--create",
        hidden=True,
    ),
    info: Optional[bool] = typer.Option(
        None,
        "-i",
        "--info",
        "--current-info",
        hidden=True,
    ),
    extra_args: Optional[List[str]] = typer.Argument(
        default=None, hidden=True, metavar=""
    ),
) -> None:
    if version:
        console.print(
            f"anaconda-auth, version [cyan]{__version__}[/cyan]",
            style="bold green",
        )
        raise typer.Exit()

    # We have to manually handle subcommands due the the handling of the auth subcommand
    # as a top-level subcommand in anaconda-client
    extra_args = extra_args or []
    if extra_args:
        subcommand_name = extra_args[0]
    else:
        subcommand_name = None

    # Extract the subcommands attached to the app. Use dynamic loading just to be safe,
    # because static typing shows this to be invalid.
    subcommands_dict = getattr(ctx.command, "commands", {})

    # If the subcommand is known, then we delegate to the actual functions defined in this module
    if cmd := subcommands_dict.get(subcommand_name):
        cmd.main(
            extra_args[1:], prog_name=subcommand_name, standalone_mode=False, parent=ctx
        )
        return

    has_legacy_options = any(
        value is not None
        for value in (
            name,
            organization,
            strength,
            strong,
            weak,
            url,
            max_age,
            scopes,
            out,
            list_scopes,
            list_tokens,
            remove,
            create,
            info,
        )
    )

    if has_legacy_options or subcommand_name:
        # If any of the anaconda-client options are passed, try to delegate to
        # binstar_main if it exists. Otherwise, we just exit gracefully.

        try:
            from binstar_client.scripts.cli import main as binstar_main
        except (ImportError, ModuleNotFoundError):
            return

        console.print(
            "[yellow]DeprecationWarning[/yellow]: Please use [cyan]anaconda org auth[/cyan] instead for explicit management of anaconda.org auth tokens\n"
        )
        warnings.warn(
            "Please use `anaconda org auth` instead for explicit management of anaconda.org auth tokens",
            DeprecationWarning,
        )

        binstar_main(sys.argv[1:], allow_plugin_main=False)
        return

    # No subcommand was given, so we print help
    console.print(ctx.get_help())


@app.command("login")
def auth_login(
    force: bool = False, ssl_verify: bool = True, at: Optional[str] = None
) -> None:
    """Login"""
    _override_default_site(at)
    try:
        token_info = TokenInfo.load()
        domain = token_info.domain
        if token_info.expired:
            console.print(f"Your API key has expired, logging into {domain}")
            login(force=True, ssl_verify=ssl_verify)
            raise typer.Exit()
    except TokenNotFoundError:
        pass  # Proceed to login
    else:
        force = force or Confirm.ask(
            f"You are already logged into Anaconda ({domain}). Would you like to force a new login?",
            default=False,
        )
        if not force:
            raise typer.Exit()

    login(force=force, ssl_verify=ssl_verify)


@app.command(name="whoami")
def auth_info(at: Optional[str] = None) -> None:
    """Display information about the currently signed-in user"""
    _override_default_site(at)
    client = BaseClient()
    response = client.get("/api/account")
    response.raise_for_status()
    console.print(f"Your info ({client.config.domain}):")
    console.print_json(data=response.json(), indent=2, sort_keys=True)


@app.command(name="api-key")
def auth_key(at: Optional[str] = None) -> None:
    """Display API Key for signed-in user"""
    _override_default_site(at)
    token_info = TokenInfo.load()
    if not token_info.expired:
        print(token_info.api_key)
        return
    else:
        raise TokenExpiredError()


@app.command(name="logout")
def auth_logout(at: Optional[str] = None) -> None:
    """Logout"""
    _override_default_site(at)
    logout()



import warnings
from functools import cached_property
from typing import Any
from typing import ClassVar
from typing import Dict
from typing import List
from typing import Literal
from typing import MutableMapping
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from urllib.parse import urljoin

from pydantic import BaseModel
from pydantic import RootModel
from pydantic.fields import FieldInfo
from pydantic_settings import BaseSettings
from pydantic_settings import PydanticBaseSettingsSource

from anaconda_auth import __version__ as version
from anaconda_auth.exceptions import UnknownSiteName
from anaconda_cli_base.config import AnacondaBaseSettings
from anaconda_cli_base.config import AnacondaConfigTomlSettingsSource
from anaconda_cli_base.config import anaconda_config_path
from anaconda_cli_base.console import console


def _raise_deprecated_field_set_warning(set_fields: Dict[str, Any]) -> None:
    fields_str = ", ".join(sorted(f'"{s}"' for s in set_fields.keys()))
    warning_text = (
        "The following fields have been set using legacy environment variables "
        + "prefixed with 'ANACONDA_CLOUD_` or in the `plugins.cloud` section "
        + f"of `~/.anaconda/config.toml`: {fields_str}\n\n"
        + "Please either rename environment variables to the corresponding "
        + "`ANACONDA_AUTH_` version, or replace the `plugins.cloud` section "
        + "of the config file with `plugins.auth`."
    )
    console.print(f"[red]{warning_text}[/red]")
    warnings.warn(
        warning_text,
        DeprecationWarning,
    )


class AnacondaAuthSite(BaseModel):
    site: Optional[str] = "anaconda.com"
    preferred_token_storage: Literal["system", "anaconda-keyring"] = "anaconda-keyring"
    domain: str = "anaconda.com"
    auth_domain_override: Optional[str] = None
    api_key: Optional[str] = None
    keyring: Optional[Dict[str, Dict[str, str]]] = None
    ssl_verify: Union[bool, Literal["truststore"]] = True
    extra_headers: Optional[Union[Dict[str, str], str]] = None
    client_id: str = "b4ad7f1d-c784-46b5-a9fe-106e50441f5a"
    redirect_uri: str = "http://127.0.0.1:8000/auth/oidc"
    openid_config_path: str = "/.well-known/openid-configuration"
    oidc_request_headers: Dict[str, str] = {"User-Agent": f"anaconda-auth/{version}"}
    login_success_path: str = "/app/local-login-success"
    login_error_path: str = "/app/local-login-error"
    use_unified_repo_api_key: bool = False
    hash_hostname: bool = True
    proxy_servers: Optional[MutableMapping[str, str]] = None
    client_cert: Optional[str] = None
    client_cert_key: Optional[str] = None
    use_device_flow: bool = False
    _merged: bool = False

    @property
    def auth_domain(self) -> str:
        """The authentication domain base URL.

        Defaults to the `auth` subdomain of the main domain.

        """
        if self.auth_domain_override:
            return self.auth_domain_override
        return self.domain

    @property
    def well_known_url(self) -> str:
        """The URL from which to load the OpenID configuration."""
        return urljoin(f"https://{self.auth_domain}", self.openid_config_path)

    @property
    def login_success_url(self) -> str:
        """The location to redirect after auth flow, if successful."""
        return urljoin(f"https://{self.domain}", self.login_success_path)

    @property
    def login_error_url(self) -> str:
        """The location to redirect after auth flow, if there is an error."""
        return urljoin(f"https://{self.domain}", self.login_error_path)

    @property
    def oidc(self) -> "OpenIDConfiguration":
        """The OIDC configuration, cached as a regular instance attribute."""
        from anaconda_auth.client import BaseClient

        client = BaseClient(site=self)
        res = client.get(
            self.well_known_url,
            auth=False,  # type: ignore
        )
        res.raise_for_status()
        oidc_config = OpenIDConfiguration(**res.json())
        return self.__dict__.setdefault("_oidc", oidc_config)

    @cached_property
    def aau_token(self) -> Union[str, None]:
        # The token is cached in anaconda_anon_usage, so we can also cache here
        try:
            from anaconda_anon_usage.tokens import token_string
        except ImportError:
            return None

        try:
            return token_string()
        except Exception:
            # We don't want this to block user login in any case,
            # so let any Exceptions pass silently.
            return None


class AnacondaSettingsSource(PydanticBaseSettingsSource):
    def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> tuple[Any, str, bool]:
        # Nothing to do here. Only implement the return statement to make mypy happy
        return None, "", False


class CondaContextSettingsSource(AnacondaSettingsSource):
    def __call__(self) -> Dict[str, Any]:
        values = {}

        try:
            from conda.base.context import context

            if context.proxy_servers:
                values["proxy_servers"] = dict(context.proxy_servers)
            if context.client_ssl_cert:
                values["client_cert"] = context.client_ssl_cert
            if context.client_ssl_cert_key:
                values["client_cert_key"] = context.client_ssl_cert_key

            values["ssl_verify"] = context.ssl_verify

        except ImportError:
            pass

        return values


class AnacondaCloudSettingsSource(AnacondaSettingsSource):
    def __call__(self) -> Dict[str, Any]:
        cloud_config = AnacondaCloudConfig(raise_deprecation_warning=False)
        set_fields = cloud_config.model_dump(exclude_unset=True)
        if set_fields:
            _raise_deprecated_field_set_warning(set_fields)
        return set_fields


class AnacondaSiteSettingsSource(AnacondaSettingsSource):
    def __call__(self) -> Dict[str, Any]:
        state = self.current_state
        site_config = AnacondaAuthSitesConfig()
        all_sites = site_config.sites
        if state.get("site"):
            # If an explicit site is requested, the name must
            # be an exact match for a key in the sites data.
            site = all_sites._find_key(state["site"])
        elif state.get("domain"):
            # If a domain is requested:
            # - if a single match is found, use it
            # - if multiple matches are found, raise an error
            # - if no match is found, use the default
            try:
                site = all_sites._find_domain(state["domain"])
            except UnknownSiteName:
                return {}
        else:
            # We now fall back to default_site, which for historical
            # reasons can be a site key or a domain.
            site = all_sites._find_at(site_config.default_site)
        return all_sites.root[site].model_dump(exclude_unset=True)


class AnacondaAuthConfig(AnacondaAuthSite, AnacondaBaseSettings, plugin_name="auth"):
    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            env_settings,
            file_secret_settings,
            dotenv_settings,
            AnacondaSiteSettingsSource(cls),
            AnacondaCloudSettingsSource(cls),
            AnacondaConfigTomlSettingsSource(settings_cls, anaconda_config_path()),
            CondaContextSettingsSource(cls),
        )


class OpenIDConfiguration(BaseModel):
    authorization_endpoint: str
    token_endpoint: str
    device_authorization_endpoint: Optional[str] = None


_OLD_OIDC_REQUEST_HEADERS = {"User-Agent": f"anaconda-cloud-auth/{version}"}


class AnacondaCloudConfig(AnacondaAuthSite, AnacondaBaseSettings, plugin_name="cloud"):
    oidc_request_headers: Dict[str, str] = _OLD_OIDC_REQUEST_HEADERS

    def __init__(self, raise_deprecation_warning: bool = True, **kwargs: Any):
        if self.__class__ == "AnacondaCloudConfig" and raise_deprecation_warning:
            warnings.warn(
                "AnacondaCloudConfig is deprecated, please use AnacondaAuthConfig instead.",
                DeprecationWarning,
            )
        super().__init__(**kwargs)


class Sites(RootModel[Dict[str, AnacondaAuthSite]]):
    def _find_key(self, key: Optional[str]) -> str:
        if key in self.root:
            return key
        raise UnknownSiteName(
            f"The site name {key} has not been configured in {anaconda_config_path()}"
        )

    def _find_domain(self, domain: Optional[str]) -> str:
        matches = [
            (key, site) for key, site in self.root.items() if site.domain == domain
        ]
        if len(matches) == 1:
            return matches[0][0]
        if matches:
            mstr = ", ".join(skey for skey, _ in matches)
            raise ValueError(
                f"The domain {domain} matches more than one configured site ({mstr})"
            )
        elif domain == "anaconda.com":
            self.root[domain] = AnacondaAuthSite()
            return domain
        else:
            raise UnknownSiteName(
                f"The site or domain {domain} has not been configured in {anaconda_config_path()}"
            )

    def _find_at(self, key: Optional[str]) -> str:
        # Fuzzy match:
        # - If the key is an exact match for a site key, use it
        # - If the key matches a single site's domain, use it
        # - Otherwise, raise an exception
        try:
            return self._find_key(key)
        except UnknownSiteName:
            return self._find_domain(key)

    def __getitem__(self, key: str) -> AnacondaAuthConfig:
        return AnacondaAuthConfig(site=self._find_at(key))


class AnacondaAuthSitesConfig(AnacondaBaseSettings, plugin_name=None):
    _instance: ClassVar[Optional["AnacondaAuthSitesConfig"]] = None

    default_site: Optional[str] = None
    sites: Sites = Sites({})

    def __new__(cls) -> "AnacondaAuthSitesConfig":
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if self.default_site is None:
            if self.sites.root:
                self.default_site = next(iter(self.sites.root))
            else:
                self.default_site = "anaconda.com"
                self.sites.root["anaconda.com"] = AnacondaAuthSite()
        for key, value in self.sites.root.items():
            value.site = key

    @classmethod
    def all_sites(cls) -> List[str]:
        return list(cls().sites.root)

    @classmethod
    def load_site(cls, site: Optional[str] = None) -> AnacondaAuthSite:
        """Load the site configuration object (site=None loads default_site)"""
        config = cls()
        sstr: str = site or config.default_site or "anaconda.com"
        return config.sites[sstr]



import shlex
from uuid import UUID

import pytest
from pytest_mock import MockerFixture
from requests_mock import Mocker as RequestMocker

from ..conftest import CLIInvoker

pytest.importorskip("conda")

# ruff: noqa: E402

from anaconda_auth._conda import repo_config
from anaconda_auth.repo import OrganizationData
from anaconda_auth.repo import TokenCreateResponse
from anaconda_auth.token import TokenInfo
from anaconda_auth.token import TokenNotFoundError


@pytest.fixture(autouse=True)
def mock_do_auth_flow(mocker: MockerFixture) -> None:
    mocker.patch(
        "anaconda_auth.actions._do_auth_flow",
        return_value="test-access-token",
    )


@pytest.fixture(
    autouse=True, params=["security_subscription", "commercial_subscription"]
)
def user_has_business_subscription(
    request, requests_mock: RequestMocker, org_name: str, business_org_id: UUID
) -> None:
    requests_mock.get(
        "https://anaconda.com/api/account",
        json={
            "subscriptions": [
                {
                    "org_id": str(business_org_id),
                    "product_code": request.param,
                }
            ]
        },
    )


@pytest.fixture(autouse=True)
def repodata_json_available_with_token(
    requests_mock: RequestMocker, token_created_in_service: TokenCreateResponse
) -> None:
    requests_mock.head(
        f"https://repo.anaconda.cloud/t/{token_created_in_service.token}/repo/main/noarch/repodata.json",
        status_code=200,
    )


def test_issue_new_token_prints_success_message_via_cli(
    org_name: str,
    mocker: MockerFixture,
    capsys: pytest.CaptureFixture,
    token_exists_in_service,
    token_created_in_service,
    invoke_cli,
) -> None:
    result = invoke_cli(["token", "install", "--org", org_name], input="y\nn\n")

    expected_msg = "Your conda token has been installed and expires 2025-01-01 00:00:00. To view your token(s), you can use anaconda token list\n"
    assert result.exit_code == 0, result.stdout
    assert expected_msg in result.stdout


def test_token_list_no_tokens(
    invoke_cli: CLIInvoker, no_tokens_installed: None
) -> None:
    result = invoke_cli(["token", "list"])

    assert result.exit_code == 1
    assert (
        "No repo tokens are installed. Run anaconda token install." in result.stdout
    ), result.stdout
    assert "Aborted." in result.output


def test_token_list_has_tokens(mocker: MockerFixture, invoke_cli: CLIInvoker) -> None:
    test_repo_token = "test-repo-token"
    mock = mocker.patch(
        "anaconda_auth._conda.repo_config.read_binstar_tokens",
        return_value={repo_config.REPO_URL: test_repo_token},
    )
    result = invoke_cli(["token", "list"])

    mock.assert_called_once()

    assert result.exit_code == 0
    assert "Anaconda Repository Tokens" in result.stdout
    assert repo_config.REPO_URL in result.stdout
    assert test_repo_token in result.stdout


@pytest.mark.parametrize("option_flag", ["-o", "--org"])
def test_token_install_does_not_exist_yet(
    option_flag: str,
    org_name: str,
    token_does_not_exist_in_service: None,
    token_created_in_service: str,
    *,
    invoke_cli: CLIInvoker,
) -> None:
    result = invoke_cli(
        ["token", "install", option_flag, org_name],
        input="y\n",
    )
    assert result.exit_code == 0

    token_info = TokenInfo.load()
    repo_token = token_info.get_repo_token(org_name=org_name)
    assert repo_token == token_created_in_service.token


@pytest.mark.parametrize("option_flag", ["-o", "--org"])
def test_token_install_exists_already_accept(
    option_flag: str,
    org_name: str,
    token_exists_in_service: None,
    token_created_in_service: TokenCreateResponse,
    *,
    invoke_cli: CLIInvoker,
) -> None:
    result = invoke_cli(["token", "install", option_flag, org_name], input="y\ny\n")
    assert result.exit_code == 0, result.stdout

    token_info = TokenInfo.load()
    repo_token = token_info.get_repo_token(org_name=org_name)
    assert repo_token == token_created_in_service.token


@pytest.mark.parametrize("option_flag", ["-o", "--org"])
def test_token_install_exists_already_decline(
    option_flag: str,
    org_name: str,
    valid_api_key: TokenInfo,
    token_exists_in_service: None,
    token_created_in_service: str,
    *,
    invoke_cli: CLIInvoker,
) -> None:
    result = invoke_cli(["token", "install", option_flag, org_name], input="n")
    assert result.exit_code == 1, result.stdout

    token_info = TokenInfo.load()
    with pytest.raises(TokenNotFoundError):
        _ = token_info.get_repo_token(org_name=org_name)


def test_token_install_no_available_org(
    user_has_no_orgs: list[OrganizationData],
    *,
    invoke_cli: CLIInvoker,
) -> None:
    result = invoke_cli(["token", "install"])
    assert result.exit_code == 1, result.stdout
    assert "No organizations found." in result.stdout, result.stdout
    assert "Aborted." in result.output


def test_token_install_select_first_if_only_org(
    org_name: str,
    token_does_not_exist_in_service: None,
    token_created_in_service: str,
    user_has_one_org: list[OrganizationData],
    *,
    invoke_cli: CLIInvoker,
) -> None:
    result = invoke_cli(["token", "install"], input="y")
    assert result.exit_code == 0, result.stdout
    assert (
        f"Only one organization found, automatically selecting: {org_name}"
        in result.stdout
    )

    token_info = TokenInfo.load()
    repo_token = token_info.get_repo_token(org_name=org_name)
    assert repo_token == token_created_in_service.token


def test_token_install_select_second_of_multiple_orgs(
    org_name: str,
    token_does_not_exist_in_service: None,
    token_created_in_service: str,
    user_has_multiple_orgs: list[OrganizationData],
    *,
    invoke_cli: CLIInvoker,
) -> None:
    # TODO: This uses the "j" key binding. I can't figure out how to send the right
    #       escape code for down arrow.
    result = invoke_cli(["token", "install"], input="j\ny\n")
    assert result.exit_code == 0, result.stdout

    token_info = TokenInfo.load()
    repo_token = token_info.get_repo_token(org_name=org_name)
    assert repo_token == token_created_in_service.token


@pytest.mark.parametrize("option_flag", ["-o", "--org"])
def test_token_uninstall(
    option_flag: str,
    org_name: str,
    token_is_installed: TokenInfo,
    *,
    invoke_cli: CLIInvoker,
) -> None:
    result = invoke_cli(["token", "uninstall", option_flag, org_name])
    assert result.exit_code == 0, result.stdout

    token_info = TokenInfo.load()
    with pytest.raises(TokenNotFoundError):
        _ = token_info.get_repo_token(org_name=org_name)


@pytest.mark.parametrize("option_flag", ["-a", "--all"])
def test_token_uninstall_all(
    option_flag,
    token_is_installed: TokenInfo,
    org_name: str,
    *,
    invoke_cli: CLIInvoker,
) -> None:
    token_info = TokenInfo.load()

    result = invoke_cli(["token", "uninstall", option_flag])
    assert result.exit_code == 0, result.stdout

    token_info = TokenInfo.load()

    with pytest.raises(TokenNotFoundError):
        _ = token_info.get_repo_token(org_name=org_name)


def test_token_remove(
    token_is_installed: TokenInfo,
    org_name,
    *,
    invoke_cli: CLIInvoker,
) -> None:
    exp_dict = repo_config.token_list()
    repo_config.token_set(
        token_is_installed.get_repo_token(org_name=org_name), force=True
    )
    exp_token = token_is_installed.get_repo_token(org_name=org_name)
    exp_dict["https://repo.anaconda.cloud/repo/"] = exp_token
    assert repo_config.token_list() == exp_dict, repo_config.token_list()
    result = invoke_cli(
        [
            "token",
            "remove",
        ]
    )
    assert result.exit_code == 0, result.stdout
    del exp_dict["https://repo.anaconda.cloud/repo/"]
    assert repo_config.token_list() == exp_dict, repo_config.token_list()


@pytest.mark.parametrize(
    "option_flag",
    [
        "-e",
        "--env",
        "-f ./some/path/.condarc",
        "--file ./some/path/other/.condarc",
        "-s",
        "--system",
    ],
)
def test_set_token_without_org(
    option_flag,
    org_name: str,
    mocker: MockerFixture,
    capsys: pytest.CaptureFixture,
    token_exists_in_service,
    token_created_in_service,
    invoke_cli,
) -> None:
    result = invoke_cli(
        ["token", "set", *(shlex.split(option_flag)), token_created_in_service.token],
        input="y\nn\n",
    )

    assert result.exit_code == 0, result.stdout

    bin_token = repo_config.token_list()
    assert (
        token_created_in_service.token == bin_token["https://repo.anaconda.cloud/repo/"]
    ), bin_token
    repo_config.token_remove()



import pytest
from requests import PreparedRequest
from requests import Response
from requests.hooks import dispatch_hook

from anaconda_auth.token import TokenInfo

conda = pytest.importorskip("conda")

from conda.base import context  # noqa: E402
from conda.base.context import context as conda_context  # noqa: E402
from conda.gateways.connection.session import CondaSession  # noqa: E402
from conda.gateways.connection.session import get_session  # noqa: E402

from anaconda_auth._conda import config as plugin_config  # noqa: E402
from anaconda_auth._conda.auth_handler import AnacondaAuthError  # noqa: E402
from anaconda_auth._conda.auth_handler import AnacondaAuthHandler  # noqa: E402
from anaconda_auth._conda.condarc import CondaRC  # noqa: E402


@pytest.fixture()
def mocked_empty_conda_token(mocker):
    mocker.patch(
        "anaconda_auth._conda.repo_config.token_list",
        return_value={},
    )


@pytest.fixture()
def mocked_conda_token(mocker):
    mocker.patch(
        "anaconda_auth._conda.repo_config.token_list",
        return_value={"https://repo.anaconda.cloud/repo/": "my-test-token"},
    )


@pytest.fixture()
def mocked_token_info(mocker):
    mocker.patch(
        "anaconda_auth.token.TokenInfo.load",
        return_value=TokenInfo(
            domain="repo.anaconda.cloud",
            repo_tokens=[
                {
                    "org_name": "my-first-org",
                    "token": "my-first-test-token-in-token-info",
                },
                {"org_name": "my-org", "token": "my-test-token-in-token-info"},
            ],
        ),
    )


@pytest.fixture()
def mocked_token_info_with_api_key(mocker):
    mocker.patch(
        "anaconda_auth.token.TokenInfo.load",
        return_value=TokenInfo(
            domain="repo.anaconda.cloud",
            api_key="my-test-api-key",
            repo_tokens=[
                {
                    "org_name": "my-first-org",
                    "token": "my-first-test-token-in-token-info",
                },
                {"org_name": "my-org", "token": "my-test-token-in-token-info"},
            ],
        ),
    )


@pytest.fixture()
def handler():
    return AnacondaAuthHandler(
        channel_name="https://repo.anaconda.cloud/repo/my-org/my-channel"
    )


@pytest.mark.usefixtures("mocked_conda_token")
def test_get_token_via_conda_token(handler):
    token = handler._load_token(
        "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"
    )
    assert token == "my-test-token"


@pytest.mark.usefixtures("mocked_token_info")
def test_get_repo_token_via_keyring(handler):
    token = handler._load_token(
        "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"
    )
    assert token == "my-test-token-in-token-info"


@pytest.mark.usefixtures("mocked_token_info_with_api_key")
def test_get_unified_api_token_for_dotcom(handler, monkeypatch):
    # It should not matter what this value is; the API key should still be attached
    monkeypatch.setenv("ANACONDA_AUTH_USE_UNIFIED_REPO_API_KEY", "False")
    for host in ("repo.anaconda.com", "repo.continuum.io"):
        token = handler._load_token(f"https://{host}/pkgs/main/noarch/repodata.json")
        assert token == "my-test-api-key"


@pytest.mark.usefixtures("mocked_token_info_with_api_key")
def test_get_unified_api_token_via_keyring(handler, monkeypatch):
    monkeypatch.setenv("ANACONDA_AUTH_USE_UNIFIED_REPO_API_KEY", "True")
    token = handler._load_token(
        "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"
    )
    assert token == "my-test-api-key"


@pytest.mark.usefixtures("mocked_token_info")
def test_auth_handler_call_sets_authorization_header_repo_token(handler, monkeypatch):
    request = PreparedRequest()
    request.url = (
        "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"
    )
    request.headers = {}

    modified_request = handler(request)

    assert (
        modified_request.headers["Authorization"] == "token my-test-token-in-token-info"
    )


@pytest.mark.usefixtures("mocked_token_info_with_api_key")
def test_auth_handler_call_sets_authorization_header_unified_api_token(
    handler, monkeypatch
):
    monkeypatch.setenv("ANACONDA_AUTH_USE_UNIFIED_REPO_API_KEY", "True")

    request = PreparedRequest()
    request.url = (
        "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"
    )
    request.headers = {}

    modified_request = handler(request)

    assert modified_request.headers["Authorization"] == "Bearer my-test-api-key"


@pytest.mark.usefixtures("mocked_token_info")
def test_get_token_for_main_finds_first_token(handler):
    token = handler._load_token(
        "https://repo.anaconda.cloud/repo/main/noarch/repodata.json"
    )
    assert token == "my-first-test-token-in-token-info"


@pytest.mark.usefixtures("mocked_empty_conda_token")
def test_get_token_missing(handler):
    token = handler._load_token(
        "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"
    )
    assert token is None


@pytest.fixture()
def url() -> str:
    return "https://repo.anaconda.cloud/repo/my-org/my-channel/noarch/repodata.json"


@pytest.fixture()
def session(handler, url) -> CondaSession:
    # Create a session and assign the handler to it
    get_session.cache_clear()
    session_obj = get_session(url)
    session_obj.auth = handler
    return session_obj


@pytest.mark.usefixtures("mocked_token_info")
def test_inject_header_during_request(session, url, monkeypatch):
    # Set up a dummy function that will capture the PreparedRequest without sending it.
    request = None

    def capture_request(req, *args, **kwargs):
        nonlocal request
        request = req

    monkeypatch.setattr(session, "send", capture_request)

    # Make sure the token got injected
    session.get(url)
    assert request.headers.get("Authorization") == "token my-test-token-in-token-info"


@pytest.mark.parametrize("mocked_status_code", [401, 403])
@pytest.mark.usefixtures("mocked_token_info")
def test_response_callback_error_handler(
    mocked_status_code, *, session, url, monkeypatch
):
    def _mocked_request(req, *args, **kwargs):
        response = Response()
        response.status_code = mocked_status_code
        response = dispatch_hook("response", req.hooks, response, **kwargs)
        return response

    monkeypatch.setattr(session, "send", _mocked_request)

    # A 403 response is captured by the hook and a custom exception is raised
    with pytest.raises(AnacondaAuthError):
        session.get(url)


@pytest.mark.parametrize("mocked_status_code", [401, 403])
def test_inject_no_header_during_request_if_no_token(
    mocked_status_code, *, session, url, monkeypatch
):
    """
    If there is not token, we first make a request without an Authorization header.
    If the server responds with an error code, we raise an exception.
    """
    # Set up a dummy function that will capture the PreparedRequest without sending it.
    request = None

    def _mocked_request(req, *args, **kwargs):
        # Capture the request object for introspection later
        nonlocal request
        request = req

        # Simulate a 403 response from the server
        response = Response()
        response.status_code = mocked_status_code
        response = dispatch_hook("response", req.hooks, response, **kwargs)
        return response

    monkeypatch.setattr(session, "send", _mocked_request)

    # An error response is captured by the hook and a custom exception is raised
    with pytest.raises(AnacondaAuthError):
        session.get(url)

    # Make sure the token did not get injected
    assert request.headers.get("Authorization") is None


REFERENCE = {
    "https://repo.continuum.io/*": "anaconda-auth",
    "https://repo.anaconda.com/*": "anaconda-auth",
    "https://repo.anaconda.cloud/*": "anaconda-auth",
    "https://anaconda.com/*": "anaconda-auth",
}


def test_channel_settings_yaml():
    yaml1 = plugin_config._build_channel_yaml(False, False).splitlines()
    assert "# DO NOT EDIT THIS FILE." in yaml1
    assert "channel_settings: []" in yaml1
    yaml2 = plugin_config._build_channel_yaml(True, False).splitlines()
    assert "# DO NOT EDIT THIS FILE." in yaml2
    assert "channel_settings:" in yaml2
    assert any(line[0] in "#c- " for line in yaml2)


def test_channel_settings_empty(conda_search_path):
    plugin_config._assert_settings(conda_context, {})


def test_channel_settings_prefix(conda_search_path):
    fpath = conda_search_path.prefix / "anaconda-auth.yml"
    assert not fpath.exists()
    plugin_config._write_condarc_d_settings()
    with pytest.raises(FileExistsError):
        plugin_config._write_condarc_d_settings()
    plugin_config._write_condarc_d_settings(overwrite=True)
    plugin_config._verify_channel_settings(filtered=False)
    context.reset_context()
    plugin_config._assert_settings(conda_context, REFERENCE)


def test_channel_settings_user(conda_search_path):
    fpath = conda_search_path.user
    assert not fpath.read_text().strip()
    condarc = CondaRC(fpath)
    condarc.update_channel_settings("my-test-channel", "anaconda-auth", username=None)
    condarc.save()
    context.reset_context()
    plugin_config._assert_settings(conda_context, {"my-test-channel": "anaconda-auth"})


def test_channel_settings_site(conda_search_path):
    condarc = CondaRC(conda_search_path.sites / "anaconda-auth-sites.yml")
    condarc.update_channel_settings("my-site-channel", "anaconda-auth", username=None)
    condarc.save()
    context.reset_context()
    plugin_config._assert_settings(conda_context, {"my-site-channel": "anaconda-auth"})


def test_channel_settings_merged(conda_search_path):
    plugin_config._write_condarc_d_settings()

    condarc1 = CondaRC(conda_search_path.user)
    condarc1.update_channel_settings("my-test-channel", "anaconda-auth", username=None)
    condarc1.save()

    condarc2 = CondaRC(conda_search_path.sites / "anaconda-auth-sites.yml")
    condarc2.update_channel_settings("my-site-channel", "anaconda-auth", username=None)
    condarc2.save()

    context.reset_context()
    expected = REFERENCE.copy()
    expected["my-test-channel"] = "anaconda-auth"
    expected["my-site-channel"] = "anaconda-auth"
    plugin_config._assert_settings(conda_context, expected)



from textwrap import dedent

import pytest
from packaging.version import parse

from anaconda_auth._conda.repo_config import CONDA_VERSION
from anaconda_auth._conda.repo_config import _set_ssl_verify_false
from anaconda_auth._conda.repo_config import configure_default_channels
from anaconda_auth._conda.repo_config import enable_extra_safety_checks


def test_default_channels(condarc_path):
    final_condarc = dedent(
        """\
        default_channels:
          - https://repo.anaconda.cloud/repo/main
          - https://repo.anaconda.cloud/repo/r
          - https://repo.anaconda.cloud/repo/msys2
        """
    )
    configure_default_channels(force=True)
    assert condarc_path.read_text() == final_condarc


def test_default_channels_no_exception(condarc_path, capsys):
    """Ensure that no CondaKeyError is raised if the .condarc does not have default_channels defined."""
    configure_default_channels(force=True)

    res = capsys.readouterr()
    assert "CondaKeyError: 'default_channels'" not in res.err


def test_replace_default_channels(condarc_path):
    original_condarc = dedent(
        """\
        default_channels:
          - https://repo.anaconda.com/pkg/main
          - https://repo.anaconda.com/pkg/r
          - https://repo.anaconda.com/pkg/msys2
        """
    )
    final_condarc = dedent(
        """\
        default_channels:
          - https://repo.anaconda.cloud/repo/main
          - https://repo.anaconda.cloud/repo/r
          - https://repo.anaconda.cloud/repo/msys2
        """
    )
    condarc_path.write_text(original_condarc)
    configure_default_channels(force=True)
    assert condarc_path.read_text() == final_condarc


def test_default_channels_with_inactive(condarc_path):
    original_condarc = dedent(
        """\
        default_channels:
          - https://repo.anaconda.com/pkg/main
          - https://repo.anaconda.com/pkg/r
          - https://repo.anaconda.com/pkg/msys2
        """
    )
    final_condarc = dedent(
        """\
        default_channels:
          - https://repo.anaconda.cloud/repo/main
          - https://repo.anaconda.cloud/repo/r
          - https://repo.anaconda.cloud/repo/msys2
          - https://repo.anaconda.cloud/repo/free
          - https://repo.anaconda.cloud/repo/pro
          - https://repo.anaconda.cloud/repo/mro-archive
        """
    )
    condarc_path.write_text(original_condarc)
    configure_default_channels(
        include_archive_channels=["free", "pro", "mro-archive"], force=True
    )
    assert condarc_path.read_text() == final_condarc


def test_replace_default_channels_with_inactive(condarc_path):
    final_condarc = dedent(
        """\
        default_channels:
          - https://repo.anaconda.cloud/repo/main
          - https://repo.anaconda.cloud/repo/r
          - https://repo.anaconda.cloud/repo/msys2
          - https://repo.anaconda.cloud/repo/free
          - https://repo.anaconda.cloud/repo/pro
          - https://repo.anaconda.cloud/repo/mro-archive
        """
    )
    configure_default_channels(
        include_archive_channels=["free", "pro", "mro-archive"], force=True
    )
    assert condarc_path.read_text() == final_condarc


def test_default_channels_with_conda_forge(condarc_path):
    original_condarc = dedent(
        """\
        ssl_verify: true

        default_channels:
          - https://repo.anaconda.com/pkgs/main
        channels:
          - defaults
          - conda-forge

        channel_alias: https://conda.anaconda.org/
        """
    )

    condarc_path.write_text(original_condarc)
    configure_default_channels(force=True)
    assert condarc_path.read_text() == dedent(
        """\
        ssl_verify: true

        channels:
          - defaults
          - conda-forge

        channel_alias: https://conda.anaconda.org/
        default_channels:
          - https://repo.anaconda.cloud/repo/main
          - https://repo.anaconda.cloud/repo/r
          - https://repo.anaconda.cloud/repo/msys2
        """
    )


def test_no_ssl_verify_from_true(condarc_path):
    original_condarc = dedent(
        """\
        ssl_verify: true
        """
    )
    final_condarc = dedent(
        """\
        ssl_verify: false
        """
    )
    condarc_path.write_text(original_condarc)
    _set_ssl_verify_false()
    assert condarc_path.read_text() == final_condarc


def test_no_ssl_verify_from_empty(condarc_path):
    final_condarc = dedent(
        """\
        ssl_verify: false
        """
    )
    _set_ssl_verify_false()
    assert condarc_path.read_text() == final_condarc


def test_no_ssl_verify_from_false(condarc_path):
    original_condarc = dedent(
        """\
        ssl_verify: false
        """
    )
    final_condarc = dedent(
        """\
        ssl_verify: false
        """
    )

    condarc_path.write_text(original_condarc)
    _set_ssl_verify_false()
    assert condarc_path.read_text() == final_condarc


@pytest.mark.skipif(
    CONDA_VERSION < parse("4.10.1"),
    reason="Signature verification was added in Conda 4.10.1",
)
def test_enable_package_signing(condarc_path):
    final_condarc = dedent(
        """\
        extra_safety_checks: true
        signing_metadata_url_base: https://repo.anaconda.cloud/repo
        """
    )

    enable_extra_safety_checks()
    assert condarc_path.read_text() == final_condarc



"""Configuration manager for the conda plugin.

This file is used as a central location to manage global configuration
settings and deliver them to the locations where they are needed, including
other modules and YAML configuration files. It is designed to be callable as
a standalone script or as a module. It can install a conda configuration file
in the prefix, and to verify its existence and expected behavioral impact.
These functions are both used by the conda recipe itself.

It is imperative that this module not include any other anaconda_auth imports
so that it can be run during the conda build process
"""

import sys
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Set
from typing import Union

__all__ = []


PREFIX_CONDARC_PATH = Path(sys.prefix) / "condarc.d" / "anaconda-auth.yml"


# This list is now serving THREE purposes. The keys are used in the conda
# plugin module to determine which hosts should be hardcoded to use
# anaconda-auth for authentication. The values are used to provide the
# keyring domain where the legacy token will be stored, as well as
# whether or not the destination should receive a proper API key.
# Finally, the list is used to generate the master anaconda-auth.yml
# configuration for conda.
class TokenDomainSetting(NamedTuple):
    token_domain: str
    default_use_unified_api_key: bool = True


TOKEN_DOMAIN_MAP = {
    "repo.continuum.io": TokenDomainSetting("anaconda.com"),
    "repo.anaconda.com": TokenDomainSetting("anaconda.com"),
    "repo.anaconda.cloud": TokenDomainSetting("anaconda.com", False),
}


def _dictify(channel_settings: List[Dict[str, str]]) -> Dict[str, str]:
    """
    Extract the authentication information for each channel
    for better comparison with expectations.
    """
    return {c["channel"]: c.get("auth", "") for c in channel_settings}


def _assert_settings(
    context: Any, expected: Dict[str, str], filtered: bool = False
) -> None:
    """
    Compare just the channel authentication information for each
    channel with the expectation, with order independence
    """
    found = _dictify(context.channel_settings)
    if filtered:
        found = {k: v for k, v in expected.items() if k in expected}
    assert found == expected


def _build_channel_settings(
    include_defaults: bool = True, include_sites: bool = True
) -> List[Dict[str, str]]:
    hosts: Set[str] = set()
    if include_defaults:
        hosts.update(TOKEN_DOMAIN_MAP)
        hosts.update(t.token_domain for t in TOKEN_DOMAIN_MAP.values())
    if include_sites:
        # We are delaying this import so this file can be run standalone
        from anaconda_auth.config import AnacondaAuthSitesConfig

        hosts.update(s.domain for s in AnacondaAuthSitesConfig().sites.root.values())
    return [{"channel": f"https://{host}/*", "auth": "anaconda-auth"} for host in hosts]


def _build_channel_yaml(
    include_defaults: bool = True, include_sites: bool = True
) -> str:
    settings = _build_channel_settings(include_defaults, include_sites)
    # Hand-construct the YAML so we can avoid a library dependency, and
    # easily include the warning comment.
    lines = [
        "# DO NOT EDIT THIS FILE.",
        "# This file was generated automatically for anaconda-auth.",
    ]
    lines.append("channel_settings:")
    for record in settings:
        pfx = "- "
        for key, value in record.items():
            lines.append(f'{pfx}{key}: "{value}"')
            pfx = "  "
    if not settings:
        lines[-1] += " []"
    lines.append("")
    return "\n".join(lines)


def _write_channel_settings(
    fpath: Union[Path, str],
    include_defaults: bool = True,
    include_sites: bool = True,
    overwrite: bool = False,
) -> None:
    with Path(fpath).open(mode="w" if overwrite else "x") as fp:
        fp.write(_build_channel_yaml(include_defaults, include_sites))


def _write_condarc_d_settings(overwrite: bool = False) -> None:
    PREFIX_CONDARC_PATH.parent.mkdir(parents=True, exist_ok=True)
    _write_channel_settings(
        PREFIX_CONDARC_PATH, include_sites=False, overwrite=overwrite
    )


def _verify_channel_settings(filtered: bool = True) -> None:
    from conda.base.context import context

    context.__init__()
    expected = _build_channel_settings(include_sites=False)
    _assert_settings(context, _dictify(expected), filtered=filtered)


if __name__ == "__main__":
    import sys

    invalid = [
        x for x in sys.argv[1:] if x not in ("--install", "--verify", "--overwrite")
    ]
    if len(sys.argv) == 1 or invalid:
        print(
            f"Usage: {sys.argv[0]} [--install] [--verify] [--overwrite]",
            file=sys.stderr,
        )
        if invalid:
            print(f"Invalid option(s): {' '.join(invalid)}", file=sys.stderr)
        sys.exit(1)
    if "--install" in sys.argv:
        _write_condarc_d_settings(overwrite="--overwrite" in sys.argv)
    if "--verify" in sys.argv:
        _verify_channel_settings()
